#!/usr/bin/env python3

import binascii
import socketserver
import signal
import time
import socket
import string
from hashlib import md5
from base64 import b64decode
from base64 import b64encode

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

HOST = "127.0.0.1"
PORT = 4444
SOCK = None
ENCRYPTED_MSG = None
BLOCK_SIZE = 16

# ------------------------------------------------------------

def query_oracle(payload):
    global SOCK
    encoded_payload = b64encode(payload)
    SOCK.send(encoded_payload)
    oracle_reply = SOCK.recv(4096)
    return b"OK" in oracle_reply

# ---------------

def decrypt_block(encrypted_text, n, block_size=8):
    """ Decrypts the n-th block of the encrypted_text """
    assert n > 1, "Cannot decrypt first block as we don't know the IV"

    # are we the last block? this matter because the last block is the
    # only block that is properly padded, and therefore could cause a
    # false positive answer when guessing the last byte with the value
    # of 0x1.
    # 
    # Since then the xor would eliminate, the ciphertext woulnd't
    # change and therefore we would simply submit the block as it is,
    # getting a trivial OK answer.
    last_block = (len(encrypted_text) // block_size) == n

    # To decrypt block n we need to change bytes of the previous
    # block, which is block n - 1.
    saved_bytes = bytes(encrypted_text[block_size * (n-2): block_size * (n-1)])
    guess_so_far = [0] * block_size
    for i in range(0, block_size):
        msg = list(encrypted_text[:block_size * n])
        padding = i + 1
        found = False
        for c in range(0, 256):
            
            # skip this case, as it would lead us to false
            # positives, since the last block is the only block
            # that is properly padded.            
            if last_block and i == 0 and c == 1:
                continue
            
            guess_so_far[block_size - 1 - i] = c
            
            # prepare payload
            for j in range(0, i + 1):
                global_index = block_size * (n - 1) - j - 1
                relative_index = block_size - j - 1
                msg[global_index] = saved_bytes[relative_index] ^ guess_so_far[relative_index] ^ padding
                
            # test payload
            r = query_oracle(bytes(msg))

            # if we found something, then we stop trying the remaining
            # bytes for the current index and we go to the next.
            if r:
                found = True
                if chr(c) in string.printable:
                    print(f"[{n}]: byte {block_size - i} is: {c}, {chr(c)}")
                else:
                    print(f"[{n}]: byte {block_size - i} is: {c}, non-printable byte")
                break

        # since we have not found any other alternatives, we can
        # safely assume that this is the correct value
        if not found and last_block and i == 0:
            guess_so_far[block_size - 1 - i] = 1

    return guess_so_far

def challenge_3_solution_main():
    global HOST, PORT, SOCK, ENCRYPTED_MSG, BLOCK_SIZE

    SOCK = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    SOCK.settimeout(300)
    SOCK.connect((HOST, PORT))

    server_data = SOCK.recv(4096)
    start_delimiter = b"\n\nENCRYPTED_FLAG WITH CBC-AES: "
    end_delimiter = b"\n\n> "
    i = server_data.find(start_delimiter) + len(start_delimiter)
    j = server_data.find(end_delimiter)

    ENCRYPTED_MSG = server_data[i:j]

    # given that we don't know the IV we can only decrypt starting from the 2nd block.
    encrypted_msg_bytes = b64decode(ENCRYPTED_MSG)
    num_blocks = int(len(encrypted_msg_bytes) / BLOCK_SIZE)

    plaintext = []
    for i in range(2, num_blocks + 1):
        plaintext += decrypt_block(encrypted_msg_bytes, i, block_size=BLOCK_SIZE)
    plaintext = "".join(map(lambda x : chr(x), plaintext))
    print(f"Plaintext obtained is: {plaintext}")    
    SOCK.close()

if __name__ == "__main__":
    challenge_3_solution_main()
